# -*- coding: utf-8 -*-

from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ObjectDoesNotExist
from django import forms
from django.contrib.auth.models import User
from django.contrib.auth import get_backends

from .models import Bookmark, UserProfile
from tcms.core.forms import StripURLField


IM_CHOICES = (
    (1, 'IRC'),
    (2, 'Jabber'),
    (3, 'MSN'),
    (4, 'Yahoo messenger'),
    (5, 'ICQ')
)

BOOKMARK_EMPTY_LABEL = '---all---'


class UserProfileForm(forms.ModelForm):
    user = forms.CharField(widget=forms.HiddenInput)
    username = forms.RegexField(
        label=_("Username"), max_length=30, regex=r'^[\w.@+-]+$',
        help_text=_(
            "Required. 30 characters or fewer. "
            "Letters, digits and @/./+/-/_ only."),
        error_messages={'invalid': _(
            "This value may contain only letters, "
            "numbers and @/./+/-/_ characters.")},
    )
    first_name = forms.CharField(max_length=128, required=False)
    last_name = forms.CharField(max_length=128, required=False)
    email = forms.EmailField(label=_("E-mail"), max_length=75)
    im_type_id = forms.ChoiceField(choices=IM_CHOICES)
    url = StripURLField(required=False)

    class Meta:
        model = UserProfile
        fields = '__all__'

    def __init__(self, *args, **kwargs):
        if 'url' in kwargs:
            kwargs['url'] = kwargs['url'].strip()
        super().__init__(*args, **kwargs)
        if 'instance' in kwargs:
            instance = kwargs['instance']
            self.initial['username'] = instance.user.username
            self.initial['first_name'] = instance.user.first_name
            self.initial['last_name'] = instance.user.last_name
            self.initial['email'] = instance.user.email

    def clean_email(self):
        email = self.cleaned_data['email']
        if not getattr(self.instance, 'pk'):
            return email

        if email == self.instance.user.email:
            return email

        try:
            u = User.objects.get(email=email)
        except ObjectDoesNotExist:
            return email

        if u == self.instance:
            return u.email

        raise forms.ValidationError(
            _("A user with that email already exists."))

    def clean_user(self):
        if not self.instance.pk:
            return User.objects.get(pk=self.cleaned_data['user'])

        if self.instance.user.pk == int(self.cleaned_data['user']):
            return self.instance.user

        raise forms.ValidationError(_("User error."))

    def clean_username(self):
        username = self.cleaned_data['username']

        if not getattr(self.instance, 'pk'):
            return username

        if username == self.instance.user.username:
            return username

        try:
            u = User.objects.get(username=username)
        except ObjectDoesNotExist:
            return username

        if u == self.instance:
            return username

        raise forms.ValidationError(
            _("A user with that username already exists."))

    def save(self, commit=True):
        can_register = False

        for b in get_backends():
            if getattr(b, 'can_register', None):
                can_register = True

        instance = super().save(commit=commit)
        user = instance.user
        if can_register:
            user.username = self.cleaned_data['username']
            user.email = self.cleaned_data['email']

        user.first_name = self.cleaned_data['first_name']
        user.last_name = self.cleaned_data['last_name']
        user.save()
        return instance


class BookmarkForm(forms.Form):
    a = forms.CharField(widget=forms.HiddenInput)
    content_type = forms.CharField(
        required=False, widget=forms.HiddenInput
    )
    object_pk = forms.CharField(
        required=False, widget=forms.HiddenInput
    )
    user = forms.IntegerField(widget=forms.HiddenInput)
    url = StripURLField()
    name = forms.CharField(max_length=1024, required=False)
    description = forms.CharField(required=False, widget=forms.Textarea)

    def clean(self):
        from django.apps import apps
        from django.conf import settings
        from django.core.exceptions import ObjectDoesNotExist, ValidationError
        from django.contrib.sites.models import Site
        from django.contrib.auth.models import User
        from django.contrib.contenttypes.models import ContentType

        cleaned_data = self.cleaned_data.copy()
        if cleaned_data.get('content_type'):
            try:
                m = apps.get_model(*cleaned_data['content_type'].split(".", 1))
                target = m._default_manager.get(pk=cleaned_data['object_pk'])
                app_label, model = cleaned_data['content_type'].split(".", 1)
                ct = ContentType.objects.get(
                    app_label=app_label, model=model
                )
                cleaned_data['content_type'] = ct
                cleaned_data['object_pk'] = target.pk
            except ObjectDoesNotExist as error:
                raise ValidationError(error)

        cleaned_data['user'] = User.objects.get(pk=cleaned_data['user'])
        cleaned_data['site'] = Site.objects.get(pk=settings.SITE_ID)
        return cleaned_data

    def populate(self, user):
        pass

    def save(self):
        cleaned_data = self.cleaned_data.copy()
        del cleaned_data['a']
        if not cleaned_data['content_type']:
            del cleaned_data['content_type']
            del cleaned_data['object_pk']
        return Bookmark.objects.create(**cleaned_data)



#!/usr/bin/env python3
#
# Nitrate is a test case management system.
# Copyright (C) 2019  Nitrate Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import argparse
import logging
import os
import re
import subprocess

from typing import Dict, List

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)

DB_CONTAINER_NAME = 'nitrate-test-db'
TEST_DB_NAME = 'nitrate'
TEST_BOX_IMAGE = 'quay.io/nitrate/testbox:latest'
VALID_NITRATE_DB_NAMES = ['mysql', 'mariadb', 'postgres', 'sqlite']
# Since this script was written originally to work inside Travis-CI, using
# Python version 3.6 and 3.7 would be much easier to match the value of
# environment variable TRAVIS_PYTHON_VERSION.
VALID_PYTHON_VERSIONS = ['3.6', '3.7']
DB_CONTAINER_INFO = {
    'mysql': {
        'db_engine': 'mysql',
        'db_image': 'mysql:8.0.20',
    },
    'mariadb': {
        'db_engine': 'mysql',
        'db_image': 'mariadb:10.4.12',
    },
    'sqlite': {
        'db_engine': 'sqlite',
        'db_image': '',
    },
    'postgres': {
        'db_engine': 'pgsql',
        'db_image': 'postgres:12.2',
    },
}


def validate_django_ver(value):
    regex = r'^django(>|>=|<|<=)[0-9]+\.[0-9]+,(>|>=|<|<=)[0-9]+\.[0-9]+$'
    if not re.match(regex, value):
        raise argparse.ArgumentTypeError(
            f"Invalid django version specifier '{value}'.")
    return value


def validate_project_dir(value):
    if os.path.exists(value):
        return value
    return argparse.ArgumentTypeError(
        'Invalid project root directory. It might not exist.')


def docker_run(image,
               rm: bool = False,
               detach: bool = False,
               interactive: bool = False,
               tty: bool = False,
               name: str = None,
               link: str = None,
               volumes: List[str] = None,
               envs: Dict[str, str] = None,
               cmd_args: List[str] = None
               ) -> None:
    cmd = ['docker', 'run']
    if rm:
        cmd.append('--rm')
    if detach:
        cmd.append('--detach')
    if interactive:
        cmd.append('-i')
    if tty:
        cmd.append('-t')
    if name:
        cmd.append('--name')
        cmd.append(name)
    if link:
        cmd.append('--link')
        cmd.append(link)
    if volumes:
        for item in volumes:
            cmd.append('--volume')
            cmd.append(item)
    if envs:
        for var_name, var_value in envs.items():
            cmd.append('--env')
            cmd.append(f'{var_name}={var_value}')
    cmd.append(image)
    if cmd_args:
        cmd.extend(cmd_args)

    log.debug('Run: %r', cmd)
    subprocess.check_call(cmd)


def docker_ps(all_: bool = False,
              filter_: List[str] = None,
              quiet: bool = False) -> str:
    cmd = ['docker', 'ps']
    if all_:
        cmd.append('--all')
    if filter_:
        for item in filter_:
            cmd.append('--filter')
            cmd.append(item)
    if quiet:
        cmd.append('--quiet')

    log.debug('Run: %r', cmd)
    return subprocess.check_output(cmd, universal_newlines=True)


def docker_stop(name: str) -> None:
    cmd = ['docker', 'stop', name]
    log.debug('Run: %r', cmd)
    subprocess.check_call(cmd)


def stop_container(name: str) -> None:
    c_hash = docker_ps(all_=True, filter_=[f'name={name}'], quiet=True)
    if c_hash:
        docker_stop(name)


def main():
    parser = argparse.ArgumentParser(
        description='Run tests matrix inside containers. This is particularly '
                    'useful for running tests in Travis-CI.'
    )
    parser.add_argument(
        '--python-ver',
        choices=VALID_PYTHON_VERSIONS,
        default='3.7',
        help='Specify Python version')
    parser.add_argument(
        '--django-ver',
        type=validate_django_ver,
        default='django<2.3,>=2.2',
        help='Specify django version specifier')
    parser.add_argument(
        '--nitrate-db',
        choices=VALID_NITRATE_DB_NAMES,
        default='sqlite',
        help='Database engine name')
    parser.add_argument(
        '--project-dir',
        metavar='DIR',
        type=validate_project_dir,
        default=os.path.abspath(os.curdir),
        help='Project root directory. Default to current working directory')
    parser.add_argument(
        'targets', nargs='+', help='Test targets')

    args = parser.parse_args()

    container_info = DB_CONTAINER_INFO[args.nitrate_db]
    db_engine = container_info['db_engine']
    db_image = container_info['db_image']

    stop_container(DB_CONTAINER_NAME)

    test_box_run_opts = None

    if db_engine == 'mysql':
        docker_run(
            db_image,
            rm=True,
            name=DB_CONTAINER_NAME,
            detach=True,
            envs={
               'MYSQL_ALLOW_EMPTY_PASSWORD': 'yes',
               'MYSQL_DATABASE': 'nitrate'
            },
            cmd_args=[
               '--character-set-server=utf8mb4',
               '--collation-server=utf8mb4_unicode_ci'
            ])
        test_box_run_opts = {
            'link': f'{DB_CONTAINER_NAME}:mysql',
            'envs': {
                'NITRATE_DB_ENGINE': db_engine,
                'NITRATE_DB_NAME': TEST_DB_NAME,
                'NITRATE_DB_HOST': DB_CONTAINER_NAME,
            }
        }
    elif db_engine == 'pgsql':
        docker_run(
            db_image,
            rm=True,
            detach=True,
            name=DB_CONTAINER_NAME,
            envs={'POSTGRES_PASSWORD': 'admin'}
        )
        test_box_run_opts = {
            'link': f'{DB_CONTAINER_NAME}:postgres',
            'envs': {
                'NITRATE_DB_ENGINE': db_engine,
                'NITRATE_DB_HOST': DB_CONTAINER_NAME,
                'NITRATE_DB_NAME': TEST_DB_NAME,
                'NITRATE_DB_USER': 'postgres',
                'NITRATE_DB_PASSWORD': 'admin',
            }
        }
    elif db_engine == 'sqlite':
        # No need to launch a SQLite docker image
        test_box_run_opts = {
            'envs': {
                'NITRATE_DB_ENGINE': db_engine,
                'NITRATE_DB_NAME': "file::memory:",
            }
        }

    test_box_container_name = f'nitrate-testbox-py{args.python_ver.replace(".", "")}'
    test_box_run_opts.update({
        'rm': True,
        'interactive': True,
        'tty': True,
        'name': test_box_container_name,
        'volumes': [f'{args.project_dir}:/code:Z'],
    })
    test_box_run_opts['envs'].update({
        'PYTHON_VER': f'py{args.python_ver.replace(".", "")}',
        'DJANGO_VER': args.django_ver,
        'TEST_TARGETS': '"{}"'.format(' '.join(args.targets)),
    })

    try:
        log.debug('Start testbox to run tests')
        docker_run(TEST_BOX_IMAGE, **test_box_run_opts)
    finally:
        log.debug('Stop container: %s', DB_CONTAINER_NAME)
        stop_container(DB_CONTAINER_NAME)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from datetime import datetime
from django.db.models import Max
from django.urls import reverse

from tcms.management.models import Product
from tcms.management.models import TestBuild
from tcms.testruns.models import TestCaseRunStatus
from tests import BaseCaseRun


class TestProductOverview(BaseCaseRun):
    """Test /report/product/$id/overview/"""

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        tcrs_get = TestCaseRunStatus.objects.get
        cls.status_idle = tcrs_get(name='IDLE')
        cls.status_running = tcrs_get(name='RUNNING')
        cls.status_passed = tcrs_get(name='PASSED')
        cls.status_failed = tcrs_get(name='FAILED')

        # Test data:
        # Runs: total(2), running(1), finished(1)
        # Case Runs: total(6), idle(2), running(1), passed(2), failed(1)

        cls.test_run_1.stop_date = datetime.now()
        cls.test_run_1.save()

        cls.case_run_1.case_run_status = cls.status_running
        cls.case_run_4.case_run_status = cls.status_passed
        cls.case_run_5.case_run_status = cls.status_passed
        cls.case_run_6.case_run_status = cls.status_failed

        cls.case_run_1.save()
        cls.case_run_4.save()
        cls.case_run_5.save()
        cls.case_run_6.save()

        cls.product_overview_url = reverse(
            'report-overview', args=[cls.product.pk])

    def test_404_if_specified_product_does_not_exist(self):
        qs = Product.objects.aggregate(max_pk=Max('pk'))
        url = reverse('report-overview', args=[qs['max_pk'] + 1])
        self.assert404(self.client.get(url))

    def test_show_overview(self):
        resp = self.client.get(self.product_overview_url)

        case_runs_subtotal = (
            # Runs
            ('Finished', 1, 50.0),
            ('Running', 1, 50.0),

            # Case runs
            ('PASSED', 2, 33.3),
            ('FAILED', 1, 16.7),
            ('IDLE', 2, 33.3),
            ('ERROR', 0, 0),
            ('PAUSED', 0, 0),
            ('BLOCKED', 0, 0),
            ('RUNNING', 1, 16.7),
            ('WAIVED', 0, 0),
        )

        for item_name, subtotal_count, percent in case_runs_subtotal:
            self.assertContains(
                resp,
                '<tr><td>{}</td><td>{}</td><td>{}%</td></tr>'.format(
                    item_name, subtotal_count, percent),
                html=True)


class TestOverallProductByBuilds(BaseCaseRun):
    """Test /report/product/$id/build/"""

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.url = reverse('report-overall-product-build', args=[cls.product.pk])

        # Test data:
        #
        # Product(1), associated TestBuild(2): cls.build and cls.product.build
        # Runs(2): cls.build has 2, cls.product.build has 0
        #   Both of runs are in running
        #   Case runs: 6
        #   FAILED case runs: 2
        #   RUNNING case runs: 1
        #   PASSED case runs: 1
        #   Rest 2 case runs: IDLE

        tcrs_get = TestCaseRunStatus.objects.get
        cls.status_running = tcrs_get(name='RUNNING')
        cls.status_passed = tcrs_get(name='PASSED')
        cls.status_failed = tcrs_get(name='FAILED')

        for case_run, status in ((cls.case_run_1, cls.status_failed),
                                 (cls.case_run_2, cls.status_failed),
                                 (cls.case_run_4, cls.status_running),
                                 (cls.case_run_6, cls.status_passed)):
            case_run.case_run_status = status
            case_run.save()

    def test_404_if_specified_product_does_not_exist(self):
        qs = Product.objects.aggregate(max_pk=Max('pk'))
        url = reverse('report-overall-product-build', args=[qs['max_pk'] + 1])
        self.assert404(self.client.get(url))

    def test_404_if_query_nonexisting_build_for_details(self):
        qs = TestBuild.objects.aggregate(max_pk=Max('pk'))
        self.assert404(self.client.get(self.url, data={'build_id': qs['max_pk'] + 1}))

    def test_by_builds(self):
        resp = self.client.get(self.url)

        stats = (
            (f'<a href="?build_id={self.build.pk}">{self.build.name}</a>',
             '<p>0/2</p>',
             '50.0',
             '<div class="strong emphasize">2 Failed</div>'),
            ('<a href="?build_id={}">{}</a>'.format(
                self.product.build.get(name='unspecified').pk, 'unspecified'),
             '<p>0/0</p>',
             '0',
             ''),
        )

        for col_build, col_runs, col_case_runs, col_failed_case_runs in stats:
            self.assertContains(
                resp,
                '''\
<tr>
    <td class="variety_0">{0}</td>
    <td class="variety_1">{1}</td>
    <td class="variety_4">
        <div class="progress-bar">
            <div class="progress-inner" style="width:{2}%">
                <div class="percent">{2}%</div>
            </div>
        </div>
    </td>
    <td>{3}</td>
</tr>
'''.format(col_build, col_runs, col_case_runs, col_failed_case_runs),
                html=True)

    def test_show_case_runs_subtotal_by_one_build(self):
        resp = self.client.get(self.url, data={'build_id': self.build.pk})

        data = (
            ('BLOCKED', 0, 0),
            ('ERROR', 0, 0),
            ('FAILED', 2, 33.3),
            ('IDLE', 2, 33.3),
            ('PASSED', 1, 16.7),
            ('PAUSED', 0, 0),
            ('RUNNING', 1, 16.7),
            ('WAIVED', 0, 0),
        )

        for status_name, count, percent in data:
            self.assertContains(
                resp,
                '<tr class="variety_{}">'
                '<td>{}</td><td>{}</td><td>{}%</td></tr>'.format(
                    status_name.lower(), status_name, count, percent
                ),
                html=True)



# -*- coding: utf-8 -*-

from django import http
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.urls import reverse
from django.db.models import Q
from django.shortcuts import render
from django.views.decorators.csrf import csrf_protect
from django.views.decorators.http import require_http_methods
from django.views.decorators.http import require_GET
from django.shortcuts import get_object_or_404

from tcms.core.raw_sql import RawSQL
from tcms.core.responses import JsonResponseBadRequest
from tcms.testplans.models import TestPlan
from tcms.testruns.models import TestRun
from tcms.profiles.models import Bookmark
from tcms.profiles.models import UserProfile
from tcms.profiles.forms import BookmarkForm, UserProfileForm


MODULE_NAME = 'profile'


@require_http_methods(['GET', 'POST'])
@login_required
def bookmark(request, username, template_name='profile/bookmarks.html'):
    """
    Bookmarks for the user
    """

    if username != request.user.username:
        return http.HttpResponseRedirect(reverse('nitrate-login'))
    else:
        up = {'user': request.user}

    class BookmarkActions:

        def add(self):
            form = BookmarkForm(request.GET)
            if not form.is_valid():
                return JsonResponseBadRequest({
                    'message': form.errors.as_text()
                })

            form.save()
            return http.JsonResponse({})

        def remove(self):
            pks = request.POST.getlist('pk')
            bks = Bookmark.objects.filter(pk__in=pks, user=request.user)
            bks.delete()

            return http.JsonResponse({})

        def render(self):
            if request.GET.get('category'):
                bks = Bookmark.objects.filter(user=request.user,
                                              category_id=request.GET['category'])
            else:
                bks = Bookmark.objects.filter(user=request.user)

            context_data = {
                'user_profile': up,
                'bookmarks': bks,
            }
            return render(request, template_name, context=context_data)

        def render_form(self):
            query = request.GET.copy()
            query['a'] = 'add'
            form = BookmarkForm(initial=query)
            form.populate(user=request.user)
            return http.HttpResponse(form.as_p())

    action = BookmarkActions()
    request_data = request.GET or request.POST
    func = getattr(action, request_data.get('a', 'render'))
    return func()


@require_http_methods(['GET', 'POST'])
@login_required
@csrf_protect
def profile(request, username, template_name='profile/info.html'):
    """Edit the profiles of the user"""
    u = get_object_or_404(User, username=username)

    try:
        up = UserProfile.get_user_profile(u)
    except ObjectDoesNotExist:
        up = UserProfile.objects.create(user=u)
    message = None
    form = UserProfileForm(instance=up)
    if request.method == 'POST':
        form = UserProfileForm(request.POST, instance=up)
        if form.is_valid():
            form.save()
            message = 'Information successfully updated.'
    context_data = {
        'user_profile': up,
        'form': form,
        'message': message,
    }
    return render(request, template_name, context=context_data)


@require_GET
@login_required
def recent(request, username, template_name='profile/recent.html'):
    """List the recent plan/run"""

    if username != request.user.username:
        return http.HttpResponseRedirect(reverse('nitrate-login'))
    else:
        up = {'user': request.user}

    runs_query = {
        'people': request.user,
        'is_active': True,
        'status': 'running',
    }

    tps = TestPlan.objects.filter(Q(author=request.user) | Q(owner=request.user))
    tps = tps.order_by('-plan_id')
    tps = tps.select_related('product', 'type')
    tps = tps.extra(select={
        'num_runs': RawSQL.num_runs,
    })
    tps_active = tps.filter(is_active=True)
    trs = TestRun.list(runs_query)
    latest_fifteen_testruns = trs.order_by('-run_id')[:15]
    test_plans_disable_count = tps.count() - tps_active.count()

    context_data = {
        'module': MODULE_NAME,
        'user_profile': up,
        'test_plans_count': tps.count(),
        'test_plans_disable_count': test_plans_disable_count,
        'test_runs_count': trs.count(),
        'last_15_test_plans': tps_active[:15],
        'last_15_test_runs': latest_fifteen_testruns,
    }
    return render(request, template_name, context=context_data)


@login_required
def redirect_to_profile(request):
    return http.HttpResponseRedirect(
        reverse('user-recent', args=[request.user.username]))



# -*- coding: utf-8 -*-


def order_targets(target, queryset, queries):
    """
    Designed to work with advance search module.
    Ordering queryset of testplan, testcase, or testrun.

    Each kind of objects, plan, case and run, are ordered by created date by
    default if ``order_by`` is missing from argument ``queries``.

    :param str target: what kind of objects to order. Valid choices are
        ``plan``, ``run`` and ``case``.
    :param queryset: the queryset of objects to be ordered.
    :type queryset: QuerySet
    :param dict queries: the ``Form.cleaned_data``.
    :return: ordered queryset.
    :rtype: QuerySet.
    """
    order_options = {
        'plan': {
            'default_order_by': 'create_date',
            'orderer': order_plan_queryset,
        },
        'run': {
            'default_order_by': 'create_date',
            'orderer': order_run_queryset,
        },
        'case': {
            'default_order_by': 'create_date',
            'orderer': order_case_queryset,
        },
    }
    orderer = order_options[target]['orderer']
    default_order_by = order_options[target]['default_order_by']
    order_by = queries.get('order_by', default_order_by)
    asc = bool(queries.get('asc', None))
    ordered_set = orderer(queryset, order_by, asc)
    return ordered_set


def order_plan_queryset(plans, field, asc=False):
    """
    Annotate the TestPlan queryset by calling order_by on it.
    """
    orderable_fields = (
        'plan_id', 'name', 'author__username', 'owner__username',
        'create_date', 'product__name', 'type',
        'num_cases', 'num_runs', 'num_children',
    )
    if field in orderable_fields:
        order_by = field
        if not asc:
            order_by = '-%s' % order_by
        plans = plans.order_by(order_by)
    return plans


def order_run_queryset(runs, field, asc=False):
    """
    Annotate the TestRun queryset by calling order_by on it.
    """
    orderable_fields = (
        'run_id', 'summary', 'manager__username',
        'default_tester__username', 'env_groups',
        'build__product__name', 'product_version',
        'plan__name'
    )
    if field in orderable_fields:
        order_by = field
        if not asc:
            order_by = '-%s' % order_by
        runs = runs.order_by(order_by)
    return runs


def order_case_queryset(cases, field, asc=False):
    """
    Annotate the TestCase queryset by calling order_by on it.
    """
    orderable_fields = (
        'case_id', 'summary', 'author__username',
        'default_tester__username', 'priority',
        'is_automated', 'category__name', 'case_status',
        'create_date'
    )
    if field in orderable_fields:
        order_by = field
        if not asc:
            order_by = '-%s' % order_by
        cases = cases.order_by(order_by)
    return cases



# -*- coding: utf-8 -*-

"""
Advance search implementations
"""

import time
from typing import List

from django.db.models.query import QuerySet
from django.http import HttpRequest
from django.shortcuts import render
from django.views.decorators.http import require_GET
from django.core.paginator import Paginator
from django.core.paginator import PageNotAnInteger

from tcms.core.raw_sql import RawSQL
from tcms.management.models import Priority
from tcms.management.models import Product
from tcms.search.forms import CaseForm, RunForm, PlanForm
from tcms.search.order import order_targets
from tcms.search.query import SmartDjangoQuery
from tcms.testcases.models import TestCase
from tcms.testplans.models import TestPlan
from tcms.testplans.models import TestPlanType
from tcms.testruns.models import TestRun


@require_GET
def advance_search(request, tmpl='search/advanced_search.html'):
    """View of /advance-search/"""
    errors = None
    data = request.GET
    target = data.get('target')
    plan_form = PlanForm(data)
    case_form = CaseForm(data)
    run_form = RunForm(data)
    # Update MultipleModelChoiceField on each form dynamically
    plan_form.populate(data)
    case_form.populate(data)
    run_form.populate(data)
    all_forms = (plan_form, case_form, run_form)
    errors = [f.errors for f in all_forms if not f.is_valid()]
    if errors or not data:
        products = Product.objects.order_by('pk').only('pk', 'name')
        plan_types = TestPlanType.objects.order_by('name').only('pk', 'name')
        priorities = Priority.objects.filter(is_active=True).order_by('value')
        errors = fmt_errors(errors)
        return render(request, tmpl, context=locals())

    start_time = time.time()
    results = query(request,
                    plan_form.cleaned_data,
                    run_form.cleaned_data,
                    case_form.cleaned_data,
                    target)
    results = order_targets(target, results, data)
    queries = fmt_queries(*[f.cleaned_data for f in all_forms])
    queries['Target'] = target
    return render_results(request, results, start_time, queries)


def query(request, plan_query, run_query, case_query, target, using='orm'):
    USING = {
        'orm': {
            'query': SmartDjangoQuery,
            'sum': sum_orm_queries
        }
    }
    Query = USING[using]['query']
    Sum = USING[using]['sum']
    plans = Query(plan_query, TestPlan.__name__)
    runs = Query(run_query, TestRun.__name__)
    cases = Query(case_query, TestCase.__name__)
    results = Sum(plans, cases, runs, target)
    return results


def sum_orm_queries(plans, cases, runs, target):
    """Search target objects together with selected relatives

    :return: a QuerySet object representing queried target objects.
    :rtype: QuerySet
    """
    plans = plans.evaluate()
    cases = cases.evaluate()
    runs = runs.evaluate()
    if target == 'run':
        if not plans and not cases:
            if runs is None:
                runs = TestRun.objects.none()
        if runs is None:
            runs = TestRun.objects.all()
        if cases:
            runs = runs.filter(case_run__case__in=cases).distinct()
        if plans:
            runs = runs.filter(plan__in=plans).distinct()
        return runs
    if target == 'plan':
        if not cases and not runs:
            if plans is None:
                plans = TestPlan.objects.none()
        if plans is None:
            plans = TestPlan.objects.all()
        if cases:
            plans = plans.filter(case__in=cases).distinct()
        if runs:
            plans = plans.filter(run__in=runs).distinct()
        plans = plans.extra(select={
            'num_cases': RawSQL.num_cases,
            'num_runs': RawSQL.num_runs,
            'num_children': RawSQL.num_plans,
        })
        return plans
    if target == 'case':
        if not plans and not runs:
            if cases is None:
                cases = TestCase.objects.none()
        if cases is None:
            cases = TestCase.objects.all()
        if runs:
            cases = cases.filter(case_run__run__in=runs).distinct()
        if plans:
            cases = cases.filter(plan__in=plans).distinct()
        return cases


def render_results(request, results, start_time, queries,
                   tmpl='search/results.html'):
    """Using a SQL "in" query and PKs as the arguments"""
    klasses = {
        'plan': {'class': TestPlan, 'result_key': 'test_plans'},
        'case': {'class': TestCase, 'result_key': 'test_cases'},
        'run': {'class': TestRun, 'result_key': 'test_runs'}
    }
    asc = bool(request.GET.get('asc', None))
    navigate_url = remove_from_request_path(request, 'page')
    query_url = remove_from_request_path(request, 'order_by')
    if asc:
        query_url = remove_from_request_path(query_url, 'asc')
    else:
        query_url = '%s&asc=True' % query_url

    paginator = Paginator(results, 20)
    page = request.GET.get('page')
    try:
        this_page = paginator.page(page)
    except PageNotAnInteger:
        this_page = paginator.page(1)

    page_start = paginator.per_page * (this_page.number - 1) + 1
    page_end = (
        page_start + min(len(this_page.object_list), paginator.per_page) - 1
    )

    calculate_associated_data(list(this_page), queries['Target'])

    end_time = time.time()
    time_cost = round(end_time - start_time, 3)
    context_data = {
        'search_target': klasses[request.GET['target']]['result_key'],
        'time_cost': time_cost,
        'queries': queries,
        'query_url': query_url,
        # For navigation
        'navigate_url': navigate_url,
        'this_page': this_page,
        'page_start': page_start,
        'page_end': page_end,
    }
    return render(request, tmpl, context=context_data)


def remove_from_request_path(request, name):
    """
    Remove a parameter from request.get_full_path() and return the modified
    path afterwards.
    """
    if isinstance(request, HttpRequest):
        path = request.get_full_path()
    else:
        path = request
    path = path.split('?')
    if len(path) > 1:
        path = path[1].split('&')
    else:
        return None
    path = [p for p in path if not p.startswith(name)]
    path = '&'.join(path)
    return '?' + path


def fmt_errors(form_errors):
    """
    Format errors collected in a Django Form for a better appearance.
    """
    errors = []
    for error in form_errors:
        for k, v in error.items():
            k = k.replace('p_product', 'product')
            k = k.replace('p_', 'product ')
            k = k.replace('cs_', 'case ')
            k = k.replace('pl_', 'plan ')
            k = k.replace('r_', 'run ')
            k = k.replace('_', ' ')
            if isinstance(v, list):
                v = ', '.join(map(str, v))
            errors.append((k, v))
    return errors


def fmt_queries(*queries):
    """Format the queries string."""
    results = {}
    for query in queries:
        for k, v in query.items():
            k = k.replace('p_product', 'product')
            k = k.replace('p_', 'product ')
            k = k.replace('cs_', 'case ')
            k = k.replace('pl_', 'plan ')
            k = k.replace('r_', 'run ')
            k = k.replace('_', ' ')
            if isinstance(v, bool) or v:
                if isinstance(v, QuerySet):
                    try:
                        v = ', '.join(o.name for o in v)
                    except AttributeError:
                        try:
                            v = ', '.join(o.value for o in v)
                        except AttributeError:
                            v = ', '.join(v)
                if isinstance(v, list):
                    v = ', '.join(map(str, v))
                results[k] = v
    return results


def calculate_associated_data(queryset: List, query_target):
    """Calculate associated data and attach to objects in queryset"""

    # FIXME: Maybe plan and case associated data could be calculated here as well.

    if query_target == 'run':
        for run in queryset:
            env_groups = run.plan.env_group.values_list('name', flat=True)
            if env_groups:
                run.associated_data = {'env_group': env_groups[0]}
            else:
                run.associated_data = {'env_group': None}



# -*- coding: utf-8 -*-

from django.urls import path

from .. import views

urlpatterns = [
    path('', views.all, name='runs-all'),
    path('ajax/', views.ajax_search, name='runs-ajax-search'),
    path('env_value/', views.env_value, name='runs-env-value'),
    path('clone/', views.clone, name='runs-clone'),
]



# -*- coding: utf-8 -*-

from django.urls import path
from .. import views

urlpatterns = [
    path('', views.all, name='plans-all'),
    path('new/', views.CreateNewPlanView.as_view(), name='plans-new'),
    path('ajax/', views.ajax_search, name='plans-ajax-search'),
    path('treeview/', views.tree_view, name='plans-treeview'),
    path('clone/', views.clone, name='plans-clone'),
    path('printable/', views.printable, name='plans-printable'),
    path('export/', views.export, name='plans-export'),
    # path('component/', views.component, name='plans-component'),

    path('component/', views.PlanComponentsActionView.as_view(),
         name='plans-component-actions'),
]



# -*- coding: utf-8 -*-


from datetime import datetime, timedelta

from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.urls import reverse
from django.db import models
from django.db.models import Q
from django.db.models.signals import post_save, post_delete, pre_save
from django.db.models import Count

from django_comments.models import Comment

from tcms.core.utils import EnumLike
from tcms.linkreference.models import LinkReference
from tcms.core.models.fields import DurationField
from tcms.core.models import TCMSActionModel
from tcms.core.tcms_router import connection
from tcms.core.utils import format_timedelta
from tcms.testcases.models import TestCaseText, NoneText
from tcms.testruns import signals as run_watchers
from tcms.issuetracker.models import Issue


try:
    from tcms.core.contrib.plugins_support.signals import register_model
except ImportError:
    register_model = None


class TestRun(TCMSActionModel):
    # Attribute names to get testrun statistics
    PERCENTAGES = (
        'failed_case_run_percent',
        'passed_case_run_percent',
        'completed_case_run_percent')

    run_id = models.AutoField(primary_key=True)
    plan_text_version = models.IntegerField()
    start_date = models.DateTimeField(auto_now_add=True, db_index=True)
    stop_date = models.DateTimeField(null=True, blank=True, db_index=True)
    summary = models.TextField()
    notes = models.TextField(blank=True)
    estimated_time = DurationField(default=timedelta(seconds=0))

    product_version = models.ForeignKey(
        'management.Version', related_name='version_run', on_delete=models.CASCADE)
    plan = models.ForeignKey(
        'testplans.TestPlan', related_name='run', on_delete=models.CASCADE)
    environment_id = models.IntegerField(default=0)
    build = models.ForeignKey(
        'management.TestBuild', related_name='build_runs', on_delete=models.CASCADE)
    manager = models.ForeignKey(
        'auth.User', related_name='manager', on_delete=models.CASCADE)
    default_tester = models.ForeignKey(
        'auth.User', null=True, blank=True,
        related_name='default_tester', on_delete=models.SET_NULL)

    env_value = models.ManyToManyField('management.TCMSEnvValue',
                                       through='testruns.TCMSEnvRunValueMap')

    tag = models.ManyToManyField('management.TestTag',
                                 through='testruns.TestRunTag')

    cc = models.ManyToManyField('auth.User', through='testruns.TestRunCC')
    auto_update_run_status = models.BooleanField(default=False)

    class Meta:
        db_table = 'test_runs'
        unique_together = ('run_id', 'product_version', 'plan_text_version')

    def __str__(self):
        return self.summary

    @classmethod
    def to_xmlrpc(cls, query=None):
        from tcms.xmlrpc.serializer import TestRunXMLRPCSerializer
        from tcms.xmlrpc.utils import distinct_filter

        _query = query or {}
        qs = distinct_filter(TestRun, _query).order_by('pk')
        s = TestRunXMLRPCSerializer(model_class=cls, queryset=qs)
        return s.serialize_queryset()

    @classmethod
    def list(cls, query):
        conditions = []

        mapping = {
            'search': lambda value: Q(run_id__icontains=value) | Q(summary__icontains=value),
            'summary': lambda value: Q(summary__icontains=value),
            'product': lambda value: Q(build__product=value),
            'product_version': lambda value: Q(product_version=value),
            'plan': lambda value: Q(plan__plan_id=int(value)) if value.isdigit() else Q(plan__name__icontains=value),
            'build': lambda value: Q(build=value),
            'env_group': lambda value: Q(plan__env_group=value),
            'people_id': lambda value: Q(manager__id=value) | Q(default_tester__id=value),
            'manager': lambda value: Q(manager=value),
            'default_tester': lambda value: Q(default_tester=value),
            'tag__name__in': lambda value: Q(tag__name__in=value),
            'case_run__assignee': lambda value: Q(case_run__assignee=value),
            'status': lambda value: {
                'running': Q(stop_date__isnull=True),
                'finished': Q(stop_date__isnull=False),
            }[value.lower()],
            'people': lambda value: {
                'default_tester': Q(default_tester=value),
                'manager': Q(manager=value),
                'people': Q(manager=value) | Q(default_tester=value),
                # TODO: Remove first one after upgrade to newer version.
                # query.set can return either '' or None sometimes, so
                # currently keeping these two lines here is a workaround.
                '': Q(manager=value) | Q(default_tester=value),
                None: Q(manager=value) | Q(default_tester=value),
            }[query.get('people_type')],
        }

        conditions = [mapping[key](value) for key, value in query.items()
                      if value and key in mapping]

        runs = cls.objects.filter(*conditions)

        value = query.get('sortby')
        if value:
            runs = runs.order_by(value)

        return runs.distinct()

    def belong_to(self, user):
        if self.manager == user or self.plan.author == user:
            return True

        return False

    def clear_estimated_time(self):
        """Converts a integer to time"""
        return format_timedelta(self.estimated_time)

    def check_all_case_runs(self, case_run_id=None):
        tcrs = self.case_run.all()
        tcrs = tcrs.select_related('case_run_status')

        for tcr in tcrs:
            if not tcr.finished():
                return False

        return True

    def get_absolute_url(self):
        return reverse('run-get', args=[self.pk])

    def get_notify_addrs(self):
        """
        Get the all related mails from the run
        """
        to = [self.manager.email]
        to.extend(self.cc.values_list('email', flat=True))
        if self.default_tester_id:
            to.append(self.default_tester.email)

        for tcr in self.case_run.select_related('assignee').all():
            if tcr.assignee_id:
                to.append(tcr.assignee.email)
        return list(set(to))

    # FIXME: rewrite to use multiple values INSERT statement
    def add_case_run(self, case, case_run_status=1, assignee=None,
                     case_text_version=None, build=None,
                     notes=None, sortkey=0):
        _case_text_version = case_text_version
        if not _case_text_version:
            _case_text_version = case.latest_text(
                text_required=False).case_text_version

        _assignee = (
            assignee or
            (case.default_tester_id and case.default_tester) or
            (self.default_tester_id and self.default_tester)
        )

        if isinstance(case_run_status, int):
            _case_run_status = TestCaseRunStatus.objects.get(id=case_run_status)
        else:
            _case_run_status = case_run_status

        return self.case_run.create(case=case,
                                    assignee=_assignee,
                                    tested_by=None,
                                    case_run_status=_case_run_status,
                                    case_text_version=_case_text_version,
                                    build=build or self.build,
                                    notes=notes,
                                    sortkey=sortkey,
                                    environment_id=self.environment_id,
                                    running_date=None,
                                    close_date=None)

    def add_tag(self, tag):
        return TestRunTag.objects.get_or_create(
            run=self,
            tag=tag
        )

    def add_cc(self, user):
        return TestRunCC.objects.get_or_create(
            run=self,
            user=user,
        )

    def add_env_value(self, env_value):
        return TCMSEnvRunValueMap.objects.get_or_create(run=self,
                                                        value=env_value)

    def remove_tag(self, tag):
        cursor = connection.writer_cursor
        cursor.execute(
            "DELETE from test_run_tags WHERE run_id = %s AND tag_id = %s",
            (self.pk, tag.pk))

    def remove_cc(self, user):
        cursor = connection.writer_cursor
        cursor.execute(
            "DELETE from test_run_cc WHERE run_id = %s AND who = %s",
            (self.run_id, user.id))

    def remove_env_value(self, env_value):
        run_env_value = TCMSEnvRunValueMap.objects.get(
            run=self,
            value=env_value,
        )
        run_env_value.delete()

    def mail(self, template, subject, context, to=[], request=None):
        from tcms.core.mailto import mailto

        to = self.get_notify_addrs()
        mailto(template, subject, to, context, request=request)

    def get_issues_count(self):
        """
            Return the count of distinct bug numbers recorded for
            this particular TestRun.
        """
        # note fom Django docs: A count() call performs a SELECT COUNT(*)
        # behind the scenes !!!
        return Issue.objects.filter(
            case_run__run=self
        ).values('issue_key').distinct().count()

    def get_percentage(self, count):
        case_run_count = self.total_num_caseruns
        if case_run_count == 0:
            return 0
        percent = float(count) / case_run_count * 100
        percent = round(percent, 2)
        return percent

    def _get_completed_case_run_percentage(self):
        ids = TestCaseRunStatus.completed_status_ids()
        completed_caserun = self.case_run.filter(
            case_run_status__in=ids)

        percentage = self.get_percentage(completed_caserun.count())
        return percentage

    # IIRC, this could be used inside the test runs search result to provide
    # data for progress bar.
    completed_case_run_percent = property(_get_completed_case_run_percentage)

    def _get_failed_case_run_num(self):
        failed_status_id = TestCaseRunStatus.name_to_id('FAILED')
        failed_caserun = self.case_run.filter(
            case_run_status=failed_status_id
        )
        return failed_caserun.count()

    failed_case_run_num = property(_get_failed_case_run_num)

    def _get_failed_case_run_percentage(self):
        percentage = self.get_percentage(self.failed_case_run_num)
        return percentage

    failed_case_run_percent = property(_get_failed_case_run_percentage)

    def _get_passed_case_run_num(self):
        passed_status_id = TestCaseRunStatus.name_to_id('PASSED')
        passed_caserun = self.case_run.filter(
            case_run_status=passed_status_id
        )
        return passed_caserun.count()

    passed_case_run_num = property(_get_passed_case_run_num)

    def _get_passed_case_run_percentage(self):
        percentage = self.get_percentage(self.passed_case_run_num)
        return percentage

    passed_case_run_percent = property(_get_passed_case_run_percentage)

    # FIXME: unused
    def get_status_case_run_num(self, status_name):
        status_id = TestCaseRunStatus.name_to_id(status_name)
        caserun = self.case_run.filter(
            case_run_status=status_id
        )
        return caserun.count()

    def _get_total_case_run_num(self):
        return self.case_run.count()

    total_num_caseruns = property(_get_total_case_run_num)

    def update_completion_status(self, is_auto_updated, is_finish=None):
        if is_auto_updated and self.auto_update_run_status:
            if self.completed_case_run_percent == 100.0:
                self.stop_date = datetime.now()
            else:
                self.stop_date = None
            self.save()
        if not is_auto_updated and not self.auto_update_run_status:
            if is_finish:
                self.stop_date = datetime.now()
            else:
                self.stop_date = None
            self.save()

    def subtotal_issues_by_case_run(self):
        """Return issues subtotal of this run

        :return: a mapping from case run pk to issues count that the case run
            has.
        :rtype: dict
        """
        q = Issue.objects.filter(
            case_run__run=self
        ).values('case_run').annotate(issues_count=Count('pk'))
        return {item['case_run']: item['issues_count'] for item in q}


# FIXME: replace TestCaseRunStatus' internal cache with Django's cache
# machanism


class TestCaseRunStatus(EnumLike, TCMSActionModel):
    complete_status_names = ('PASSED', 'ERROR', 'FAILED', 'WAIVED')
    failure_status_names = ('ERROR', 'FAILED')
    idle_status_names = ('IDLE',)

    id = models.AutoField(db_column='case_run_status_id', primary_key=True)
    name = models.CharField(max_length=60, blank=True, unique=True)
    sortkey = models.IntegerField(null=True, blank=True, default=0)
    description = models.TextField(null=True, blank=True)
    auto_blinddown = models.BooleanField(default=True)

    class Meta:
        db_table = 'test_case_run_status'

    def __str__(self):
        return self.name

    def finished(self):
        return self.name in self.complete_status_names

    @classmethod
    def completed_status_ids(cls):
        '''
        There are some status indicate that
        the testcaserun is completed.
        Return IDs of these statuses.
        '''
        return (cls.objects.filter(name__in=cls.complete_status_names)
                           .values_list('pk', flat=True))


class TestCaseRunManager(models.Manager):
    def get_automated_case_count(self):
        return self.filter(case__is_automated=1).count()

    def get_manual_case_count(self):
        return self.filter(case__is_automated=0).count()

    def get_both(self):
        count1 = self.get_automated_case_count()
        count2 = self.get_manual_case_count()
        return self.count() - count1 - count2


class TestCaseRun(TCMSActionModel):
    objects = TestCaseRunManager()
    case_run_id = models.AutoField(primary_key=True)
    case_text_version = models.IntegerField()
    running_date = models.DateTimeField(null=True, blank=True)
    close_date = models.DateTimeField(null=True, blank=True)
    notes = models.TextField(null=True, blank=True)
    sortkey = models.IntegerField(null=True, blank=True)
    environment_id = models.IntegerField(default=0)

    assignee = models.ForeignKey(
        'auth.User', blank=True, null=True,
        related_name='case_run_assignee', on_delete=models.SET_NULL)
    tested_by = models.ForeignKey(
        'auth.User', blank=True, null=True,
        related_name='case_run_tester', on_delete=models.SET_NULL)
    run = models.ForeignKey(
        TestRun, related_name='case_run', on_delete=models.CASCADE)
    case = models.ForeignKey(
        'testcases.TestCase', related_name='case_run', on_delete=models.CASCADE)
    case_run_status = models.ForeignKey(
        TestCaseRunStatus,
        related_name='case_runs',
        on_delete=models.CASCADE)
    build = models.ForeignKey('management.TestBuild', on_delete=models.CASCADE)

    links = GenericRelation(LinkReference, object_id_field='object_pk')
    comments = GenericRelation(Comment, object_id_field='object_pk')

    class Meta:
        db_table = 'test_case_runs'
        unique_together = ('case', 'run', 'case_text_version')

    def __str__(self):
        return f'{self.pk}: {self.case_id}'

    @classmethod
    def to_xmlrpc(cls, query={}):
        from tcms.xmlrpc.serializer import TestCaseRunXMLRPCSerializer
        from tcms.xmlrpc.utils import distinct_filter

        qs = distinct_filter(TestCaseRun, query).order_by('pk')
        s = TestCaseRunXMLRPCSerializer(model_class=cls, queryset=qs)
        return s.serialize_queryset()

    @classmethod
    def mail_scene(cls, objects, field=None, value=None, ctype=None,
                   object_pk=None):
        tr = objects[0].run
        # scence_templates format:
        # template, subject, context
        tcrs = objects.select_related()
        scence_templates = {
            'assignee': {
                'template_name': 'mail/change_case_run_assignee.txt',
                'subject': 'Assignee of run %s has been changed' % tr.run_id,
                'recipients': tr.get_notify_addrs(),
                'context': {'test_run': tr, 'test_case_runs': tcrs},
            }
        }

        return scence_templates.get(field)

    def add_issue(self, issue_key, issue_tracker,
                  summary=None, description=None, link_external_tracker=False):
        """Add an issue to this case run

        Every argument has same meaning of argument of :meth:`TestCase.add_issue`.

        :param str issue_key: issue key to add.
        :param issue_tracker: issue tracker the issue key should belong to.
        :type issue_tracker: :class:`IssueTracker`
        :param str summary: issue's summary.
        :param str description: issue's description.
        :param bool link_external_tracker: whether to add case to issue's
            external tracker in remote issue tracker.
        :return: the newly added issue.
        :rtype: :class:`Issue`
        """
        return self.case.add_issue(
            issue_key=issue_key,
            issue_tracker=issue_tracker,
            summary=summary,
            description=description,
            case_run=self,
            link_external_tracker=link_external_tracker,
        )

    def remove_issue(self, issue_key):
        """Remove issue from this case run

        :param str issue_key: issue key to remove.
        """
        self.case.remove_issue(issue_key, case_run=self)

    def is_finished(self):
        return self.case_run_status.is_finished()

    def get_issues(self):
        """Get issues added to this case run

        :return: a queryset of the issues.
        """
        return Issue.objects.filter(case_run=self)

    def get_issues_count(self):
        """Return the number of issues added to this case run

        :return: the number of issues.
        :rtype: int
        """
        return self.get_issues().values('pk').count()

    def get_text_versions(self):
        return TestCaseText.objects.filter(
            case__pk=self.case.pk
        ).values_list('case_text_version', flat=True)

    def get_text_with_version(self, case_text_version=None):
        if case_text_version:
            try:
                return TestCaseText.objects.get(
                    case__case_id=self.case_id,
                    case_text_version=case_text_version
                )
            except TestCaseText.DoesNotExist:
                return NoneText
        try:
            return TestCaseText.objects.get(
                case__case_id=self.case_id,
                case_text_version=self.case_text_version
            )
        except TestCaseText.DoesNotExist:
            return NoneText

    def get_previous_or_next(self):
        ids = list(self.run.case_run.values_list('case_run_id', flat=True))
        current_idx = ids.index(self.case_run_id)
        prev = TestCaseRun.objects.get(case_run_id=ids[current_idx - 1])
        try:
            next = TestCaseRun.objects.get(case_run_id=ids[current_idx + 1])
        except IndexError:
            next = TestCaseRun.objects.get(case_run_id=ids[0])

        return (prev, next)

    def latest_text(self):
        try:
            return TestCaseText.objects.filter(
                case__case_id=self.case_id
            ).order_by('-case_text_version')[0]
        except IndexError:
            return NoneText


class TestRunTag(models.Model):
    tag = models.ForeignKey('management.TestTag', on_delete=models.CASCADE)
    run = models.ForeignKey(TestRun, related_name='tags', on_delete=models.CASCADE)
    user = models.IntegerField(db_column='userid', default='0')

    class Meta:
        db_table = 'test_run_tags'


class TestRunCC(models.Model):
    run = models.ForeignKey(TestRun, related_name='cc_list', on_delete=models.CASCADE)
    user = models.ForeignKey('auth.User', db_column='who', on_delete=models.CASCADE)

    class Meta:
        db_table = 'test_run_cc'
        unique_together = ('run', 'user')


class TCMSEnvRunValueMap(models.Model):
    run = models.ForeignKey(TestRun, on_delete=models.CASCADE)
    value = models.ForeignKey('management.TCMSEnvValue', on_delete=models.CASCADE)

    class Meta:
        db_table = 'tcms_env_run_value_map'


# Signals handler
def _run_listen():
    post_save.connect(run_watchers.post_run_saved, sender=TestRun)
    post_save.connect(run_watchers.post_case_run_saved, sender=TestCaseRun,
                      dispatch_uid='tcms.testruns.models.TestCaseRun')
    post_delete.connect(run_watchers.post_case_run_deleted, sender=TestCaseRun,
                        dispatch_uid='tcms.testruns.models.TestCaseRun')
    pre_save.connect(run_watchers.pre_save_clean, sender=TestRun)


if settings.LISTENING_MODEL_SIGNAL:
    _run_listen()

if register_model:
    register_model(TestRun)
    register_model(TestCaseRun)
    register_model(TestRunTag)



# -*- coding: utf-8 -*-

from collections import namedtuple
from itertools import groupby
from operator import itemgetter

from django.conf import settings
from django.db.models import Count, F

from tcms.testruns.models import TestCaseRun
from tcms.testruns.models import TestCaseRunStatus


TestCaseRunStatusSubtotal = namedtuple('TestCaseRunStatusSubtotal',
                                       'StatusSubtotal '
                                       'CaseRunsTotalCount '
                                       'CompletedPercentage '
                                       'FailurePercentage')


def stats_caseruns_status(run_id, case_run_statuss):
    """Get statistics based on case runs' status

    @param run_id: id of test run from where to get statistics
    @type run_id: int
    @param case_run_statuss: iterable object containing TestCaseRunStatus
        objects representing PASS, FAIL, WAIVED, etc.
    @type case_run_statuss: iterable object
    @return: the statistics including the number of each status mapping,
        total number of case runs, complete percent, and failure percent.
    @rtype: namedtuple
    """
    rows = TestCaseRun.objects.filter(
        run=run_id
    ).values(
        'case_run_status'
    ).annotate(status_count=Count('case_run_status'))

    caserun_statuss_subtotal = {
        status.pk: [0, status] for status in case_run_statuss
    }

    for row in rows:
        status_pk = row['case_run_status']
        caserun_statuss_subtotal[status_pk][0] = row['status_count']

    complete_count = 0
    failure_count = 0
    caseruns_total_count = 0
    status_complete_names = TestCaseRunStatus.complete_status_names
    status_failure_names = TestCaseRunStatus.failure_status_names

    for status_pk, total_info in caserun_statuss_subtotal.items():
        status_caseruns_count, caserun_status = total_info
        status_name = caserun_status.name

        caseruns_total_count += status_caseruns_count

        if status_name in status_complete_names:
            complete_count += status_caseruns_count
        if status_name in status_failure_names:
            failure_count += status_caseruns_count

    # Final calculation
    complete_percent = .0
    if caseruns_total_count:
        complete_percent = complete_count * 100.0 / caseruns_total_count
    failure_percent = .0
    if complete_count:
        failure_percent = failure_count * 100.0 / complete_count

    return TestCaseRunStatusSubtotal(caserun_statuss_subtotal,
                                     caseruns_total_count,
                                     complete_percent,
                                     failure_percent)


class TestCaseRunDataMixin:
    """Data for test case runs"""

    def stats_mode_caseruns(self, case_runs):
        """Statistics from case runs mode

        @param case_runs: iteratable object to access each case run
        @type case_runs: iterable, list, tuple
        @return: mapping between mode and the count. Example return value is
            { 'manual': I, 'automated': J, 'manual_automated': N }
        @rtype: dict
        """
        manual_count = 0
        automated_count = 0
        manual_automated_count = 0

        for case_run in case_runs:
            is_automated = case_run.case.is_automated
            if is_automated == 1:
                automated_count += 1
            elif is_automated == 0:
                manual_count += 1
            else:
                manual_automated_count += 1

        return {
            'manual': manual_count,
            'automated': automated_count,
            'manual_automated': manual_automated_count,
        }

    def get_caseruns_comments(self, run_pk):
        """Get case runs' comments

        :param int run_pk: run's pk whose comments will be retrieved.
        :return: the mapping between case run id and comments
        :rtype: dict
        """
        qs = TestCaseRun.objects.filter(
            run=run_pk,
            comments__site=settings.SITE_ID,
            comments__is_public=True,
            comments__is_removed=False,
        ).annotate(
            submit_date=F('comments__submit_date'),
            comment=F('comments__comment'),
            user_name=F('comments__user_name'),
        ).values(
            'case_run_id',
            'submit_date',
            'comment',
            'user_name',
        ).order_by('pk')

        return {
            case_run_id: list(comments) for case_run_id, comments in
            groupby(qs, itemgetter('case_run_id'))
        }

    def get_summary_stats(self, case_runs):
        """Get summary statistics from case runs

        Statistics targets:
        - the number of pending test case runs, whose status is IDLE
        - the number of completed test case runs, whose status are PASSED,
          ERROR, FAILED, WAIVED

        @param case_runs: iterable object containing case runs
        @type case_runs: iterable
        @return: a mapping between statistics target and its value
        @rtype: dict
        """
        idle_count = 0
        complete_count = 0
        complete_status_names = TestCaseRunStatus.complete_status_names
        idle_status_names = TestCaseRunStatus.idle_status_names

        for case_run in case_runs:
            status_name = case_run.case_run_status.name
            if status_name in idle_status_names:
                idle_count += 1
            elif status_name in complete_status_names:
                complete_count += 1

        return {'idle': idle_count, 'complete': complete_count}



# -*- coding: utf-8 -*-

from functools import partial

from django import forms

from tcms.management.models import Product, TestBuild, Component, Version, Priority
from tcms.testcases.forms import IssueKeyField
from tcms.testcases.models import TestCaseCategory, TestCaseStatus
from tcms.testplans.models import TestPlanType

# template-functions creating form field with required = False
LooseCF = partial(forms.CharField, required=False, max_length=200)
LooseIF = partial(forms.IntegerField, required=False)
LooseIssueF = partial(IssueKeyField, required=False, max_length=200)
LooseDF = partial(forms.DateField, required=False)
LooseBF = partial(forms.BooleanField, required=False)
LooseMF = partial(forms.MultipleChoiceField, required=False, choices=())
PlanTypeF = partial(forms.ModelMultipleChoiceField, required=False,
                    queryset=TestPlanType.objects.all())
ProductF = partial(forms.ModelMultipleChoiceField, required=False,
                   queryset=Product.objects.none())
BuildF = partial(forms.ModelMultipleChoiceField, required=False,
                 queryset=TestBuild.objects.none())
CategoryF = partial(forms.ModelMultipleChoiceField, required=False,
                    queryset=TestCaseCategory.objects.none())
ComponentF = partial(forms.ModelMultipleChoiceField, required=False,
                     queryset=Component.objects.none())
VersionF = partial(forms.ModelMultipleChoiceField, required=False,
                   queryset=Version.objects.none())


def get_choice(value, _type=str, deli=','):
    """
    Used to clean a form field where multiple choices are seperated using a
    delimiter such as comma. Removing the empty value.
    """
    try:
        results = value.split(deli)
        return [_type(r.strip()) for r in results if r]
    except Exception as e:
        raise forms.ValidationError(str(e))


def get_boolean_choice(value):
    return {
        'yes': True,
        'no': False
    }.get(value, None)


class PlanForm(forms.Form):
    pl_type = PlanTypeF()
    pl_summary = LooseCF()
    pl_id = LooseCF()
    pl_authors = LooseCF()
    pl_owners = LooseCF()
    pl_tags = LooseCF()
    pl_tags_exclude = LooseBF()
    pl_active = LooseCF()
    pl_created_since = LooseDF()
    pl_created_before = LooseDF()
    pl_product = ProductF()
    pl_component = ComponentF()
    pl_version = VersionF()

    def clean_pl_active(self):
        return get_boolean_choice(self.cleaned_data['pl_active'])

    def clean_pl_id(self):
        return get_choice(self.cleaned_data['pl_id'], _type=int)

    def clean_pl_tags(self):
        return get_choice(self.cleaned_data['pl_tags'])

    def clean_pl_authors(self):
        return get_choice(self.cleaned_data['pl_authors'])

    def clean_pl_owners(self):
        return get_choice(self.cleaned_data['pl_owners'])

    def populate(self, data):
        prod_pks = data.getlist('pl_product')
        prod_pks = [k for k in prod_pks if k]
        if prod_pks:
            qs = Product.objects.filter(pk__in=prod_pks)
            self.fields['pl_product'].queryset = qs
        comp_pks = data.getlist('pl_component')
        comp_pks = [k for k in comp_pks if k]
        if comp_pks:
            qs = Component.objects.filter(pk__in=comp_pks)
            self.fields['pl_component'].queryset = qs
        ver_pks = data.getlist('pl_version')
        ver_pks = [k for k in ver_pks if k]
        if ver_pks:
            qs = Version.objects.filter(pk__in=ver_pks)
            self.fields['pl_version'].queryset = qs


class CaseForm(forms.Form):
    cs_id = LooseCF()
    cs_summary = LooseCF()
    cs_authors = LooseCF()
    cs_tester = LooseCF()
    cs_tags = LooseCF()
    cs_issues = LooseIssueF()
    cs_status = LooseMF()
    cs_priority = LooseMF()
    cs_auto = LooseCF()
    cs_proposed = LooseCF()
    cs_script = LooseCF()
    cs_created_since = LooseDF()
    cs_created_before = LooseDF()
    cs_tags_exclude = LooseBF()
    cs_product = ProductF()
    cs_component = ComponentF()
    cs_category = CategoryF()

    def clean_cs_auto(self):
        return get_boolean_choice(self.cleaned_data['cs_auto'])

    def clean_cs_proposed(self):
        return get_boolean_choice(self.cleaned_data['cs_proposed'])

    def clean_cs_id(self):
        return get_choice(self.cleaned_data['cs_id'], _type=int)

    def clean_cs_authors(self):
        return get_choice(self.cleaned_data['cs_authors'])

    def clean_cs_issues(self):
        return get_choice(self.cleaned_data['cs_issues'], int)

    def clean_cs_tags(self):
        return get_choice(self.cleaned_data['cs_tags'])

    def clean_cs_tester(self):
        return get_choice(self.cleaned_data['cs_tester'])

    def populate(self, data):
        status_choice = list(TestCaseStatus.objects.order_by('pk')
                                                   .values_list('pk', 'name'))
        self.fields['cs_status'].choices = status_choice

        priority_choice = list(Priority.objects.order_by('pk')
                                               .values_list('pk', 'value'))
        self.fields['cs_priority'].choices = priority_choice

        prod_pks = data.getlist('cs_product')
        prod_pks = [k for k in prod_pks if k]
        if prod_pks:
            qs = Product.objects.filter(pk__in=prod_pks)
            self.fields['cs_product'].queryset = qs
        comp_pks = data.getlist('cs_component')
        comp_pks = [k for k in comp_pks if k]
        if comp_pks:
            qs = Component.objects.filter(pk__in=comp_pks)
            self.fields['cs_component'].queryset = qs
        cat_pks = data.getlist('cs_category')
        cat_pks = [k for k in cat_pks if k]
        if cat_pks:
            qs = TestCaseCategory.objects.filter(pk__in=cat_pks)
            self.fields['cs_category'].queryset = qs


class RunForm(forms.Form):
    r_id = LooseCF()
    r_summary = LooseCF()
    r_manager = LooseCF()
    r_tester = LooseCF()
    r_tags = LooseCF()
    r_running = LooseCF()
    r_begin = LooseDF()
    r_finished = LooseDF()
    r_created_since = LooseDF()
    r_created_before = LooseDF()
    r_tags_exclude = LooseBF()
    r_real_tester = LooseCF()
    r_build = BuildF()
    r_product = ProductF()
    r_version = VersionF()

    def clean_r_running(self):
        return get_boolean_choice(self.cleaned_data['r_running'])

    def clean_r_id(self):
        return get_choice(self.cleaned_data['r_id'], _type=int)

    def clean_r_tags(self):
        return get_choice(self.cleaned_data['r_tags'])

    def clean_r_tester(self):
        return get_choice(self.cleaned_data['r_tester'])

    def clean_r_real_tester(self):
        return get_choice(self.cleaned_data['r_real_tester'])

    def clean_r_manager(self):
        return get_choice(self.cleaned_data['r_manager'])

    def populate(self, data):
        prod_pks = data.getlist('r_product')
        prod_pks = [k for k in prod_pks if k]
        if prod_pks:
            qs = Product.objects.filter(pk__in=prod_pks).only('name')
            self.fields['r_product'].queryset = qs
        build_pks = data.getlist('r_build')
        build_pks = [k for k in build_pks if k]
        if build_pks:
            qs = TestBuild.objects.filter(pk__in=build_pks).only('name')
            self.fields['r_build'].queryset = qs
        ver_pks = data.getlist('r_version')
        ver_pks = [k for k in ver_pks if k]
        if ver_pks:
            qs = Version.objects.filter(pk__in=ver_pks).only('value')
            self.fields['r_version'].queryset = qs
